is.wholenumber <- function(x, tol = .Machine$double.eps^0.5)  abs(x - round(x)) < tol

### load remaining gi after the first elimination step (C) ###
get_sp <- function(filename, threshold) {
  if (!file.exists(filename)) stop("File does not exist!", call.=FALSE)
  if (!file.exists("my_gi_taxid_nucl.bin")) stop("my_gi_taxid_nucl.bin does not exist!", call.=FALSE)
  if (!file.exists("my_taxid_name.bin")) stop("my_taxid_name.bin does not exist!", call.=FALSE)
  if (!file.exists("my_nodes.bin")) stop("my_nodes.bin does not exist!", call.=FALSE)
  if (!file.exists("swt100.bin")) stop("swt100.bin does not exist!", call.=FALSE)
  if (!file.exists("gdx100.bin")) stop("gdx100.bin does not exist!", call.=FALSE)
  if (!file.exists("swt250.bin")) stop("swt250.bin does not exist!", call.=FALSE)
  if (!file.exists("gdx250.bin")) stop("gdx250.bin does not exist!", call.=FALSE)
  if (!file.exists("swt500.bin")) stop("swt500.bin does not exist!", call.=FALSE)
  if (!file.exists("gdx500.bin")) stop("gdx500.bin does not exist!", call.=FALSE)
  if (!file.exists("swt1000.bin")) stop("swt1000.bin does not exist!", call.=FALSE)
  if (!file.exists("gdx1000.bin")) stop("gdx1000.bin does not exist!", call.=FALSE)
  if (!file.exists("source.dat")) stop("source.dat does not exist!", call.=FALSE)
  
  cmd1 <- paste("./taec_b1.exe", filename, threshold, sep=" ");
  system(cmd1);
  rfname1 <- "taec_b_1_out.txt";
  psp <- read.table(rfname1);
  cmd2 <- paste("rm", rfname1, sep=" ");
  system(cmd2);
  colnames(psp) <- c("gi", "hits_s", "hits_b");
  return(psp);
}

### load remaining gi after the first elimination step with bootstrapping (C) ###
get_sp_boot <- function(filename, threshold, n.bootstrap, n.cores) {
  if (!file.exists(filename)) stop("File does not exist!", call.=FALSE)
  if (!file.exists("my_gi_taxid_nucl.bin")) stop("my_gi_taxid_nucl.bin does not exist!", call.=FALSE)
  if (!file.exists("my_taxid_name.bin")) stop("my_taxid_name.bin does not exist!", call.=FALSE)
  if (!file.exists("my_nodes.bin")) stop("my_nodes.bin does not exist!", call.=FALSE)
  if (!file.exists("swt100.bin")) stop("swt100.bin does not exist!", call.=FALSE)
  if (!file.exists("gdx100.bin")) stop("gdx100.bin does not exist!", call.=FALSE)
  if (!file.exists("swt250.bin")) stop("swt250.bin does not exist!", call.=FALSE)
  if (!file.exists("gdx250.bin")) stop("gdx250.bin does not exist!", call.=FALSE)
  if (!file.exists("swt500.bin")) stop("swt500.bin does not exist!", call.=FALSE)
  if (!file.exists("gdx500.bin")) stop("gdx500.bin does not exist!", call.=FALSE)
  if (!file.exists("swt1000.bin")) stop("swt1000.bin does not exist!", call.=FALSE)
  if (!file.exists("gdx1000.bin")) stop("gdx1000.bin does not exist!", call.=FALSE)
  if (!file.exists("source.dat")) stop("source.dat does not exist!", call.=FALSE)
  
  cmd1 <- paste("./taec_bm.exe", filename, threshold, n.bootstrap, n.cores, sep=" ");
  system(cmd1);
  rfname1 <- "taec_b_m_out.txt";
  psp <- read.table(rfname1);
  cmd2 <- paste("rm", rfname1, sep=" ");
  system(cmd2);
  colnames(psp) <- c("gi", "hits_s", "hits_b", "n_zero_hits");
  return(psp)
}

### retrieve similiar gi with the corresponding similarity factor (C) ###
dyn.load("Rgetsw.so")
get_sw <- function(gi, read.length) {
  if (!is.numeric(gi)) stop("The input must be numeric!", call.=FALSE)
  if (!is.wholenumber(read.length)) stop("Invalid value!", call.=FALSE)
  lst <- .Call("Rgetsw", as.integer(gi), as.integer(read.length))
  swt <- data.frame(gi = lst[[1]], sw = lst[[2]])
  return(swt)
}
### retrieve the taxid and scientific name for gi (C) ###
dyn.load("Rgetdesc.so")
get_desc <- function(gi) {
  if (!is.numeric(gi)) stop("The input must be numeric!", call.=FALSE)
  desc <- .Call("Rgetdesc", as.integer(gi))
  return(desc)
}

### retrieve rank (C) ###
dyn.load("Rnodes.so")
get_rank_name <- function(taxid, rank) {
  if (!is.numeric(taxid) | !is.character(rank)) stop("Input type error!", call.=FALSE)
  name <- .Call("Rnodes", as.integer(taxid), as.character(rank))
  return(name)
}

### load and store gi in descending order of hits (R) ###
load.species <- function(filename, n.bootstrap=1, n.cores=1, pcutoff) {
  if (missing(pcutoff)) pcutoff <- n.bootstrap*0.05
  if (!is.wholenumber(n.bootstrap)) stop("Number of bootstrapping must be integer!", call.=FALSE)
  if (pcutoff < 0 | pcutoff > 100) stop("Invalid value!", call.=FALSE)
  if (!is.wholenumber(n.cores)) stop("Invalid value!", call.=FALSE)
  
  if (n.bootstrap < 2) {
  	predat <- get_sp(filename, threshold=0.96)
	  predat <- predat[which(predat$hits_s > 5),]
  } else {
  	if (pcutoff < 1) {
  		pcutoff <- 1
  	} else { 
  		pcutoff <- round(pcutoff, 0)
  	}
    predat <- get_sp_boot(filename, threshold=0.96, n.bootstrap, n.cores)
    predat <- predat[which(predat$n_zero_hits < pcutoff),]
  }
  predat <- predat[with(predat, order(-hits_s)),]
  return(predat)
}

### create a matrix of similarity factors (R) ###
make_swt <- function(gis, n, read.length) {  
  tmpsw <- list()
  for (i in 1:n) tmpsw[[i]] <- get_sw(gis[i], read.length)
  swt <- matrix(rep(0, n*n), nrow=n)
  colnames(swt) <- gis
  rownames(swt) <- gis
  for (i in 1:n) {
    for(j in 1:n) {
      swt[i, j] <- match(gis[j], tmpsw[[i]]$gi, nomatch = 0)
      if (swt[i, j] != 0) swt[i, j] <- tmpsw[[i]]$sw[swt[i, j]]
    }
  }
  return(swt)
}

### filter out gi with gi <= 0 (R) ###
filter <- function(swt, r, b, n) {
  new_b <- solve(t(swt[1:n, 1:n]), b[1:n])
  n_b <- length(b)
  if (n_b > n) {
    tmp_swt <- swt[1:n, (n+1):n_b]
    if (length(new_b) == 1) {
    	tmp_b <- b[(n+1):n_b] - (t(tmp_swt)*new_b)
    } else {
    	tmp_b <- b[(n+1):n_b] - (t(tmp_swt)%*%new_b)
    }
    new_b <- c(new_b, tmp_b)
  }
  return(new_b)
}

### correct the number of reads and remove gi with |gi| <= 0 (R) ###
filcor <- function(swt, r, b, n_known) {
  n_known_cutoff <- n_known*0.99
  n_known_incl <- 0
  n_incl <- 0
  while (n_known_incl < n_known_cutoff) {
    n_incl <- n_incl + 1
    n_known_incl <- n_known_incl + r[n_incl]
  }
  new_n <- length(b)
  while (new_n >= n_incl) {
    new_b <- filter(swt, r, b, n_incl)
    neg_sp_indx <- which(new_b < 0)
    n_neg_sp <- length(neg_sp_indx)
    if (n_neg_sp > 0) {
      swt <- swt[-neg_sp_indx, -neg_sp_indx]
      new_b <- new_b[-neg_sp_indx]
      b <- b[-neg_sp_indx] 
      new_n <- new_n - n_neg_sp
    }
    if (new_n <= n_incl) break
    n_b <- length(b)
    local_max_new_b_indx <- which(new_b == max(new_b[(n_incl+1):n_b]))
    if (length(local_max_new_b_indx) > 1) local_max_new_b_indx <- local_max_new_b_indx[1]
    if (local_max_new_b_indx != (n_incl+1)) {
      tmp_max_new_b <- b[local_max_new_b_indx]
      b[local_max_new_b_indx] <- b[(n_incl+1)]
      b[(n_incl+1)] <- tmp_max_new_b
      tmp_max_new_b <- new_b[local_max_new_b_indx]
      new_b[local_max_new_b_indx] <- new_b[(n_incl+1)]
      new_b[(n_incl+1)] <- tmp_max_new_b
      swt[c(n_incl+1, local_max_new_b_indx),] <- swt[c(local_max_new_b_indx, n_incl+1),]
      rownames(swt)[c(n_incl+1, local_max_new_b_indx)] <- rownames(swt)[c(local_max_new_b_indx, n_incl+1)]
      swt[,c(n_incl+1, local_max_new_b_indx)] <- swt[,c(local_max_new_b_indx, n_incl+1)]
      colnames(swt)[c(n_incl+1, local_max_new_b_indx)] <- colnames(swt)[c(local_max_new_b_indx, n_incl+1)]
    }
    n_incl <- n_incl + 1;
  }
  f_output <- data.frame(gi=as.integer(colnames(swt)), hits=round(new_b, 0))
  n_f_output <- nrow(f_output)
  taxid <- NULL
  name <- NULL
  for (i in 1:n_f_output) {
    gi_desc <- get_desc(f_output$gi[i])
    taxid <- c(taxid, gi_desc[1])
    name <- c(name, gi_desc[2])
  }
  result <- NULL
  result <- data.frame(gi=f_output$gi, taxid=as.integer(taxid), hits=f_output$hits, name=name)
  if (n_f_output > 1) {
    result <- result[with(result, order(taxid)),]
    rm_rows <- NULL
    for (i in 1:(n_f_output-1)) {
      for (j in (i+1):n_f_output) {
        if (result$taxid[i] == result$taxid[j]) {
          result$hits[i] <- result$hits[i] + result$hits[j]
          rm_rows <- c(rm_rows, j)
        } else {
        	break
        }
      }
    }
    if (length(rm_rows) > 0) result <- result[-rm_rows,]
  }
  result <- result[with(result, order(-hits)),]
  return(list(genome.composition=result, similarity.factors=swt))
}

### Taxonomical Analysis by Elimination and Correction (R) ###
taec <- function(data, read.length=250, show.swt=FALSE) {
  threshold1 <- round(data$hits_s[1] * 0.0005, 0); #min(round(data$hits_s[1] * 0.0005, 0), 10)
  dat <- data[data$hits_s > threshold1,]
  n_known <- sum(dat$hits_s)
  gis <- dat$gi
  n <- length(gis)
  swt <- matrix(rep(0, n*n), nrow=n)
  swt <- make_swt(gis, n, read.length)
  r <- dat$hits_s
  b <- dat$hits_b
  result <- filcor(swt, r, b, n_known)
  rownames(result$genome.composition) <- c(1:nrow(result$genome.composition))

  if (show.swt==FALSE){
    result$genome.composition <- result$genome.composition[,-1];
    return(result$genome.composition);
  } else{
    suspicious_g <- which(apply(result$similarity.factors, 1, function(x) length(which(x>0.4)))>1)
    if (length(suspicious_g)==0){
      result$sgsf <- NULL;
    } else{
      for(i in 1:length(suspicious_g)){
        suspicious_g <- c(suspicious_g, which(result$similarity.factors[suspicious_g[i],] %in% result$similarity.factors[suspicious_g[i], which(result$similarity.factors[suspicious_g[i],]>0.4 & result$similarity.factors[suspicious_g[i],]<1)]));
      }
      
      suspicious_g = sort(unique(suspicious_g));
      
      suspicious_g_w <- result$similarity.factors[suspicious_g, suspicious_g]
      n.suspicious_g = length(suspicious_g);
      tid = sapply(as.numeric(rownames(suspicious_g_w)), function(y) get_desc(y)[1]);
      for(i in 1:n.suspicious_g){
        for(j in 1:n.suspicious_g){
          if(suspicious_g_w[i,j]==1){
            suspicious_g_w[i,j] = result$genome.composition$hits[result$genome.composition$taxid == as.numeric(tid[j])];
          }
        }
      }
      rownames(suspicious_g_w) <- tid;
      colnames(suspicious_g_w) <- tid;
      result$sgsf <- round(suspicious_g_w, 3)
    }
    return(list(genome.composition=result$genome.composition[,-1], suspicious.genomes=result$sgsf))
  } 
}

### compute distribution at a specified rank (R) ###
taxa.composition <- function(rslt, taxon) {
  if (taxon<1 | taxon>8) stop("Taxon number must be: 1-species, 2-genus, 3-family, 4-order, 5-class, 6-phylum, 7-kingdom, 8-superkingdom\n", call.=FALSE)
  rank.name <- switch(taxon, "species", "genus", "family", "order", "class", "phylum", "kingdom", "superkingdom")
  dist <- rslt
  dist$name <- as.character(dist$name)
  n_dist <- nrow(dist)
  for (j in 1:n_dist) dist$name[j] <- get_rank_name(dist$taxid[j], as.character(rank.name))
  dist <- dist[with(dist, order(name)),]
  rm_rows <- NULL
  for (i in 1:(n_dist-1)) {
    for (j in (i+1):n_dist) {
      if (dist$name[i] == dist$name[j]) {
        dist$hits[i] <- dist$hits[i] + dist$hits[j]
        rm_rows <- c(rm_rows, j)
      } else {
      	break
      }
    }
  }
  if (length(rm_rows) > 0) dist <- dist[-rm_rows,]
  dist <- dist[with(dist, order(-hits)),]
  rownames(dist) <- c(1:nrow(dist))
  return(dist)
}
